@@ -85,6 +85,8 @@ struct load_segment {
 	uint32_t region_size;		/* Mirror */
 	unsigned clustered;		/* Mirror */
 	unsigned mirror_area_count;	/* Mirror */
+	uint32_t flags;			/* Mirror log */
+	char *uuid;			/* Clustered mirror log */
 };
 
 /* Per-device properties */
@@ -127,6 +129,7 @@ struct dm_tree {
 	struct dm_hash_table *devs;
 	struct dm_hash_table *uuids;
 	struct dm_tree_node root;
+	int skip_lockfs;		/* 1 skips lockfs (for non-snapshots) */
 };
 
 /* FIXME Consider exporting this */
@@ -139,7 +142,7 @@ static int _dm_snprintf(char *buf, size_t bufsize, const char *format, ...)
         n = vsnprintf(buf, bufsize, format, ap);
         va_end(ap);
 
-        if (n < 0 || (n > bufsize - 1))
+        if (n < 0 || (n > (int) bufsize - 1))
                 return -1;
 
         return n;
@@ -158,6 +161,7 @@ struct dm_tree *dm_tree_create(void)
 	dtree->root.dtree = dtree;
 	list_init(&dtree->root.uses);
 	list_init(&dtree->root.used_by);
+	dtree->skip_lockfs = 0;
 
 	if (!(dtree->mem = dm_pool_create("dtree", 1024))) {
 		log_error("dtree pool creation failed");
@@ -899,12 +903,13 @@ static int _resume_node(const char *name, uint32_t major, uint32_t minor,
 }
 
 static int _suspend_node(const char *name, uint32_t major, uint32_t minor,
-			 struct dm_info *newinfo)
+			 int skip_lockfs, struct dm_info *newinfo)
 {
 	struct dm_task *dmt;
 	int r;
 
-	log_verbose("Suspending %s (%" PRIu32 ":%" PRIu32 ")", name, major, minor);
+	log_verbose("Suspending %s (%" PRIu32 ":%" PRIu32 ")%s", name, major,
+		    minor, skip_lockfs ? "" : " with filesystem sync.");
 
 	if (!(dmt = dm_task_create(DM_DEVICE_SUSPEND))) {
 		log_error("Suspend dm_task creation failed for %s", name);
@@ -920,6 +925,9 @@ static int _suspend_node(const char *name, uint32_t major, uint32_t minor,
 	if (!dm_task_no_open_count(dmt))
 		log_error("Failed to disable open_count");
 
+	if (skip_lockfs && !dm_task_skip_lockfs(dmt))
+		log_error("Failed to set skip_lockfs flag.");
+
 	if ((r = dm_task_run(dmt)))
 		r = dm_task_get_info(dmt, newinfo);
 
@@ -978,6 +986,11 @@ int dm_tree_deactivate_children(struct dm_tree_node *dnode,
 	return 1;
 }
 
+void dm_tree_skip_lockfs(struct dm_tree_node *dnode)
+{
+	dnode->dtree->skip_lockfs = 1;
+}
+
 int dm_tree_suspend_children(struct dm_tree_node *dnode,
 				   const char *uuid_prefix,
 				   size_t uuid_prefix_len)
@@ -1018,7 +1031,8 @@ int dm_tree_suspend_children(struct dm_tree_node *dnode,
 		    !info.exists)
 			continue;
 
-		if (!_suspend_node(name, info.major, info.minor, &newinfo)) {
+		if (!_suspend_node(name, info.major, info.minor,
+				   child->dtree->skip_lockfs, &newinfo)) {
 			log_error("Unable to suspend %s (%" PRIu32
 				  ":%" PRIu32 ")", name, info.major,
 				  info.minor);
@@ -1211,10 +1225,12 @@ static int _emit_areas_line(struct dm_task *dmt, struct load_segment *seg, char
 
 static int _emit_segment_line(struct dm_task *dmt, struct load_segment *seg, uint64_t *seg_start, char *params, size_t paramsize)
 {
+	unsigned log_parm_count;
         int pos = 0;
 	int tw;
         int r;
 	char originbuf[10], cowbuf[10], logbuf[10];
+	const char *logtype;
 
 	switch(seg->type) {
 	case SEG_ERROR:
@@ -1223,33 +1239,84 @@ static int _emit_segment_line(struct dm_task *dmt, struct load_segment *seg, uin
 	case SEG_LINEAR:
 		break;
 	case SEG_MIRRORED:
+		log_parm_count = 1;	/* Region size */
+		log_parm_count += hweight32(seg->flags);	/* [no]sync, block_on_error etc. */
+
 		if (seg->clustered) {
-			if ((tw = _dm_snprintf(params + pos, paramsize - pos, "clustered ")) < 0) {
+			if (seg->uuid)
+				log_parm_count++;	/* uuid */
+			if ((tw = _dm_snprintf(params + pos, paramsize - pos, "clustered_")) < 0) {
                         	stack;	/* Out of space */
                         	return -1;
                 	}
 			pos += tw;
 		}
-		if (!seg->log) {
-			if ((tw = _dm_snprintf(params + pos, paramsize - pos, "core 1 ")) < 0) {
-                        	stack;	/* Out of space */
-                        	return -1;
-                	}
-			pos += tw;
-		} else {
+
+		if (!seg->log)
+			logtype = "core";
+		else {
+			logtype = "disk";
+			log_parm_count++;
 			if (!_build_dev_string(logbuf, sizeof(logbuf), seg->log))
 				return_0;
-			if ((tw = _dm_snprintf(params + pos, paramsize - pos, "disk 2 %s ", logbuf)) < 0) {
+		}
+
+		if ((tw = _dm_snprintf(params + pos, paramsize - pos, "%s %u ", logtype, log_parm_count)) < 0) {
+                        stack;	/* Out of space */
+                        return -1;
+                }
+		pos += tw;
+
+		if (seg->log) {
+			if ((tw = _dm_snprintf(params + pos, paramsize - pos, "%s ", logbuf)) < 0) {
                         	stack;	/* Out of space */
                         	return -1;
                 	}
 			pos += tw;
 		}
-		if ((tw = _dm_snprintf(params + pos, paramsize - pos, "%u %u ", seg->region_size, seg->mirror_area_count)) < 0) {
+
+		if ((tw = _dm_snprintf(params + pos, paramsize - pos, "%u ", seg->region_size)) < 0) {
                        	stack; /* Out of space */
                        	return -1;
                	}
 		pos += tw;
+
+		if (seg->clustered && seg->uuid) {
+			if ((tw = _dm_snprintf(params + pos, paramsize - pos, "%s ", seg->uuid)) < 0) {
+				stack;  /* Out of space */
+				return -1;
+			}
+			pos += tw;
+		}
+
+		if ((seg->flags & DM_NOSYNC)) {
+			if ((tw = _dm_snprintf(params + pos, paramsize - pos, "nosync ")) < 0) {
+                       		stack; /* Out of space */
+                       		return -1;
+               		}
+			pos += tw;
+		} else if ((seg->flags & DM_FORCESYNC)) {
+			if ((tw = _dm_snprintf(params + pos, paramsize - pos, "sync ")) < 0) {
+                       		stack; /* Out of space */
+                       		return -1;
+               		}
+			pos += tw;
+		}
+
+		if ((seg->flags & DM_BLOCK_ON_ERROR)) {
+			if ((tw = _dm_snprintf(params + pos, paramsize - pos, "block_on_error ")) < 0) {
+                       		stack; /* Out of space */
+                       		return -1;
+               		}
+			pos += tw;
+		}
+
+		if ((tw = _dm_snprintf(params + pos, paramsize - pos, "%u ", seg->mirror_area_count)) < 0) {
+                       	stack; /* Out of space */
+                       	return -1;
+               	}
+		pos += tw;
+
 		break;
 	case SEG_SNAPSHOT:
 		if (!_build_dev_string(originbuf, sizeof(originbuf), seg->origin))
@@ -1620,7 +1687,8 @@ int dm_tree_node_add_mirror_target_log(struct dm_tree_node *node,
 					  uint32_t region_size,
 					  unsigned clustered, 
 					  const char *log_uuid,
-					  unsigned area_count)
+					  unsigned area_count,
+					  uint32_t flags)
 {
 	struct dm_tree_node *log_node = NULL;
 	struct load_segment *seg;
@@ -1637,14 +1705,21 @@ int dm_tree_node_add_mirror_target_log(struct dm_tree_node *node,
 			log_error("Couldn't find mirror log uuid %s.", log_uuid);
 			return 0;
 		}
+
 		if (!_link_tree_nodes(node, log_node))
 			return_0;
+
+		if (!(seg->uuid = dm_pool_strdup(node->dtree->mem, log_uuid))) {
+			log_error("log uuid pool_strdup failed");
+			return 0;
+		}
 	}
 
 	seg->log = log_node;
 	seg->region_size = region_size;
 	seg->clustered = clustered;
 	seg->mirror_area_count = area_count;
+	seg->flags = flags;
 
 	return 1;
 }
